/*
(C) 2000 Arjan van de Ven and others  licensed under the terms of the GPL
Modified by VDA
*/

#include <stdio.h>
#include <stdlib.h>

//Intel
#define PREFETCH " prefetchnta "
#define FEMMS ""
#define EMMS ""

//AMD
//#define PREFETCH " prefetch "
//#define FEMMS " femms\n"
//#define EMMS " emms\n"

/* This makes adding/removing test functions easier
: */
#include "normal_clear.c"
#include "repstosl_clear.c"
#include "movq_clear.c"
#include "movntq_clear.c"
#include "test_clear.c"

#include "normal_copy.c"
#include "repmovsl_copy.c"
#include "movq_copy.c"
#include "movqpf_copy.c"
#include "movqpf2_copy.c"
#include "movntq_copy.c"
#include "movntq1_copy.c"
#include "test_copy.c"
#include "npf_copy.c"

const int TRY_TIMES = 8;
const int BUFSIZE = 16 * 1024*1024;

// PAGESIZE is not tunable!
const int PAGESIZE = 4096;

static inline long long rdtsc()
{
	unsigned int low,high;
	__asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
;
	return low + (((long long)high)<<32);
}

typedef void (clear_func)(void *);
typedef void (copy_func)(void *,void *);

void test_one_clearpage(clear_func *func, char *name, char *buffer)
{
	char *temp;
	int i;
	unsigned long long before,after,min,max;

	// dummy run to ensure coherent mem/cache state
	temp = buffer;
	while(temp < buffer+BUFSIZE) {
		func(temp);
		temp += PAGESIZE;
	}
	// pick fastest run
	min = ~0ULL;
	max = 0;
	for (i=0;i<TRY_TIMES;i++) {
		temp = buffer;
		// add memset clobber here
		before = rdtsc();
    		while(temp < buffer+BUFSIZE) {
			func(temp);
			temp += PAGESIZE;
		}
		after = rdtsc();
		// add check here
		if (before>after) {
			printf("timer overflow\n");
		}
 else {
			after-=before;
			if(min>after) min=after;
			if(max<after) max=after;
		}
	}
	printf("%32s - took %5lli max,%5lli min cycles per page\n",
		name,
		max/(BUFSIZE/PAGESIZE),
		min/(BUFSIZE/PAGESIZE)
	);
}

void test_one_copypage(copy_func *func, char *name, char *buffer)
{
	char *temp;
	int i;
	unsigned long long before,after,min,max;

	// dummy run to ensure coherent mem/cache state
        temp = buffer;
	while(temp < buffer+BUFSIZE/2) {
		func(temp, temp+BUFSIZE/2);
		temp += PAGESIZE;
	}
	// pick fastest run
	min = ~0ULL;
	max = 0;
	for (i=0;i<TRY_TIMES;i++) {
;
	        temp = buffer;
		// add memset clobber here
		before = rdtsc();
		while(temp < buffer+BUFSIZE/2) {
			func(temp, temp+BUFSIZE/2);
			temp += PAGESIZE;
		}
		after = rdtsc();
		// add memcmp check here
		if (before>after) {
			printf("timer overflow\n");
		}
 else {
			after-=before;
			if(min>after) min=after;
			if(max<after) max=after;
		}
	}
	printf("%32s - took %5lli max,%5lli min cycles per page\n",
		name,
		max/(BUFSIZE/2 / PAGESIZE),
		min/(BUFSIZE/2 / PAGESIZE)
	);
}


void test_clearpage(char *buffer)
{
	printf("clear_page() tests:\n");

#define	TEST_CLEAR(a) test_one_clearpage(a,#a,buffer)
	TEST_CLEAR(normal_clear_page	);
	TEST_CLEAR(normal_clear_page	);
	TEST_CLEAR(repstosl_clear_page	);
	TEST_CLEAR(movq_clear_page	);
	TEST_CLEAR(movntq_clear_page	);
//	TEST_CLEAR(test_clear_page	);

#undef TEST_CLEAR

}

void test_copypage(char *buffer)
{
	printf("copy_page() tests:\n");

#define	TEST_COPY(a) test_one_copypage(a,#a,buffer)
	TEST_COPY(normal_copy_page	);
	TEST_COPY(normal_copy_page	);
	TEST_COPY(repmovsl_copy_page	);
	TEST_COPY(movq_copy_page	);
	TEST_COPY(movntq_copy_page	);
	TEST_COPY(movntq1_copy_page	);
	TEST_COPY(movqpf_copy_page	);
	TEST_COPY(movqpf2_copy_page	);
	TEST_COPY(npf_copy_page		);
//	TEST_COPY(test_copy_page	);

#undef TEST_COPY

}

int main()
{
	char *buffer;
	printf("Page clear/copy benchmark program.\n"
		"buffer size: %i Mb\n"
		"Each test tried %i times, max and min CPU cycles per page are reported.\n"
		"Please disregard max values. They are due to system interference only.\n",
		BUFSIZE/1024/1024,
		TRY_TIMES
	);
	buffer = malloc(BUFSIZE);
	if(!buffer)
 {
		printf("Malloc failed.\n");
		exit(1);
	}

	test_clearpage(buffer);
	printf("\n");
	test_copypage(buffer);

	printf("Done.\n");
	free(buffer);
	return 0;
}
